
import numpy as np
import pyaudio
import time
from six.moves import queue
from edge_impulse_linux.runner import ImpulseRunner as ImpulseRunner
CHUNK_SIZE = 1024
OVERLAP = 0.25

def now():
    return round(time.time() * 1000)

class Microphone():
    def __init__(self, rate, chunk_size, device_id = None, channels = 1):
        self.buff = queue.Queue()
        self.chunk_size = chunk_size
        self.data = []
        self.rate = rate
        self.closed = True
        self.channels = channels
        self.interface = pyaudio.PyAudio()
        self.device_id = device_id
        self.zero_counter = 0

        while self.device_id == None or not self.checkDeviceModelCompatibility(self.device_id):
            input_devices = self.listAvailableDevices()
            input_device_id = int(input("Type the id of the audio device you want to use: \n"))
            for device in input_devices:
                if device[0] == input_device_id:
                      if self.checkDeviceModelCompatibility(input_device_id):
                          self.device_id = input_device_id
                      else:
                          print('That device is not compatible')

        print('selected Audio device: %i'% self.device_id)

    def checkDeviceModelCompatibility(self, device_id):
        supported = False
        try:
            supported = self.interface.is_format_supported(self.rate,
                        input_device=device_id,
                        input_channels=self.channels,
                        input_format=pyaudio.paInt16)
        except:
            supported = False
        finally:
            return supported



    def listAvailableDevices(self):
        if not self.interface:
            self.interface = pyaudio.PyAudio()

        info = self.interface.get_host_api_info_by_index(0)
        numdevices = info.get('deviceCount')
        input_devices = []
        for i in range (0,numdevices):
            if self.interface.get_device_info_by_host_api_device_index(0,i).get('maxInputChannels')>0:
                input_devices.append((i, self.interface.get_device_info_by_host_api_device_index(0,i).get('name')))

        if len(input_devices) == 0:
            raise Exception('There are no audio devices available');

        for i in range (0, len(input_devices)):
            print("%i --> %s" % input_devices[i])

        return input_devices

    def __enter__(self):
        if not self.interface:
            self.interface = pyaudio.PyAudio()

        self.stream = self.interface.open(
            input_device_index = self.device_id,
            format = pyaudio.paInt16,
            channels = self.channels,
            rate = self.rate,
            input = True,
            frames_per_buffer = self.chunk_size,
            stream_callback = self.fill_buffer
        )
        self.closed = False
        return self

    def __exit__(self, type, value, traceback):
        self.stream.stop_stream()
        self.stream.close()
        self.closed = True
        self.interface.terminate()

    def fill_buffer(self, in_data, frame_count, time_info, status_flags):
        zeros=bytes(self.chunk_size*2)
        if in_data != zeros:
            self.zero_counter = 0
        else:
            self.zero_counter+=1

        if self.zero_counter > self.rate / self.chunk_size:
            self.closed = True
            raise Exception('There is no audio data comming from the audio interface')

        self.buff.put(in_data)
        return None, pyaudio.paContinue

    def generator(self):
        while not self.closed:
            chunk = self.buff.get()

            if chunk is None:
                return

            data = [chunk]
            while True:
                try:
                    chunk = self.buff.get(block=False)
                    if chunk is None:
                        return
                    data.append(chunk)
                except queue.Empty:
                    break

            yield b''.join(data)

class AudioImpulseRunner(ImpulseRunner):
    def __init__(self, model_path: str):
        super(AudioImpulseRunner, self).__init__(model_path)
        self.closed = True
        self.sampling_rate = 0
        self.window_size = 0
        self.labels = []

    def init(self):
        model_info = super(AudioImpulseRunner, self).init()
        if model_info['model_parameters']['frequency'] == 0:
            raise Exception('Model file "' + self._model_path + '" is not suitable for audio recognition')

        self.window_size = model_info['model_parameters']['input_features_count']
        self.sampling_rate = model_info['model_parameters']['frequency']
        self.labels = model_info['model_parameters']['labels']

        return model_info

    def __enter__(self):
        self.closed = False
        return self

    def __exit__(self, type, value, traceback):
        self.closed = True

    def classify(self, data):
        return super(AudioImpulseRunner, self).classify(data)

    def classifier(self, device_id = None):
        with Microphone(self.sampling_rate, CHUNK_SIZE, device_id=device_id) as mic:
            generator = mic.generator()
            features = np.array([], dtype=np.int16)
            while not self.closed:
                for audio in generator:
                    data = np.frombuffer(audio, dtype=np.int16)
                    features = np.concatenate((features, data), axis=0)
                    while len(features) >= self.window_size:
                        begin = now()
                        res = self.classify(features[:self.window_size].tolist())
                        features = features[int(self.window_size * OVERLAP):]
                        yield res, audio



import subprocess
import os.path
import tempfile
import shutil
import time
import signal
import socket
import json

def now():
    return round(time.time() * 1000)

class ImpulseRunner:
    def __init__(self, model_path: str):
        self._model_path = model_path
        self._tempdir = None
        self._runner = None
        self._client = None
        self._ix = 0

    def init(self):
        if (not os.path.exists(self._model_path)):
            raise Exception('Model file does not exist: ' + self._model_path)

        if (not os.access(self._model_path, os.X_OK)):
            raise Exception('Model file "' + self._model_path + '" is not executable')

        self._tempdir = tempfile.mkdtemp()
        socket_path = os.path.join(self._tempdir, 'runner.sock')
        self._runner = subprocess.Popen([self._model_path, socket_path], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        while not os.path.exists(socket_path) or not self._runner.poll() is None:
            time.sleep(0.1)

        if not self._runner.poll() is None:
            raise Exception('Failed to start runner (' + str(self._runner.poll()) + ')')

        self._client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self._client.connect(socket_path)

        return self.hello()

    def stop(self):
        if (self._tempdir):
            shutil.rmtree(self._tempdir)

        if (self._client):
            self._client.close()

        if (self._runner):
            os.kill(self._runner.pid, signal.SIGINT)
            # todo: in Node we send a SIGHUP after 0.5sec if process has not died, can we do this somehow here too?

    def hello(self):
        msg = { "hello": 1 }
        return self.send_msg(msg)

    def classify(self, data):
        msg = { "classify": data }
        return self.send_msg(msg)

    def send_msg(self, msg):
        t_send_msg = now()

        if not self._client:
            raise Exception('ImpulseRunner is not initialized')

        self._ix = self._ix + 1
        ix = self._ix

        msg["id"] = ix
        self._client.send(json.dumps(msg).encode('utf-8'))

        t_sent_msg = now()

        # i'm not sure if this is right, we should switch to async i/o for this like in Node
        # I think that would perform better
        data = self._client.recv(1 * 1024 * 1024)

        t_received_msg = now()

        braces_open = 0
        braces_closed = 0
        line = ''
        resp = None

        for c in data.decode('utf-8'):
            if c == '{':
                line = line + c
                braces_open = braces_open + 1
            elif c == '}':
                line = line + c
                braces_closed = braces_closed + 1
                if (braces_closed == braces_open):
                    resp = json.loads(line)
            elif braces_open > 0:
                line = line + c

            if (not resp is None):
                break

        if (not resp or resp["id"] != ix):
            raise Exception('Wrong id, expected: ' + str(ix) + ' but got ' + resp["id"])

        if not resp["success"]:
            raise Exception(resp["error"])

        del resp["id"]
        del resp["success"]

        t_parsed_msg = now()
        # print('sent', t_sent_msg - t_send_msg, 'received', t_received_msg - t_send_msg, 'parsed', t_parsed_msg - t_send_msg)
        return resp



#!/usr/bin/env python

import numpy as np
import cv2
from edge_impulse_linux.runner import ImpulseRunner
import math
import psutil

class ImageImpulseRunner(ImpulseRunner):
    def __init__(self, model_path: str):
        super(ImageImpulseRunner, self).__init__(model_path)
        self.closed = True
        self.labels = []
        self.dim = (0, 0)
        self.videoCapture = cv2.VideoCapture()
        self.isGrayscale = False

    def init(self):
        model_info = super(ImageImpulseRunner, self).init()
        width = model_info['model_parameters']['image_input_width'];
        height = model_info['model_parameters']['image_input_height'];

        if width == 0 or height == 0:
            raise Exception('Model file "' + self._model_path + '" is not suitable for image recognition')

        self.dim = (width, height)
        self.labels = model_info['model_parameters']['labels']
        self.isGrayscale =  model_info['model_parameters']['image_channel_count'] == 1
        return model_info

    def __enter__(self):
        self.videoCapture = cv2.VideoCapture()
        self.closed = False
        return self

    def __exit__(self, type, value, traceback):
        self.videoCapture.release()
        self.closed = True

    def classify(self, data):
        return super(ImageImpulseRunner, self).classify(data)

    # This returns images in RGB format (not BGR)
    def get_frames(self, videoDeviceId = 0):
        if psutil.OSX or psutil.MACOS:
            print('Make sure to grant the this script access to your webcam.')
            print('If your webcam is not responding, try running "tccutil reset Camera" to reset the camera access privileges.')

        self.videoCapture = cv2.VideoCapture(videoDeviceId)
        while not self.closed:
            success, img = self.videoCapture.read()
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            if success:
                yield img

    # This returns images in RGB format (not BGR)
    def classifier(self, videoDeviceId = 0):
        if psutil.OSX or psutil.MACOS:
            print('Make sure to grant the this script access to your webcam.')
            print('If your webcam is not responding, try running "tccutil reset Camera" to reset the camera access privileges.')

        self.videoCapture = cv2.VideoCapture(videoDeviceId)
        while not self.closed:
            success, img = self.videoCapture.read()
            if success:
                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                features, cropped = self.get_features_from_image(img)

                res = self.classify(features)
                yield res, cropped

    # This expects images in RGB format (not BGR)
    def get_features_from_image(self, img, crop_direction_x='center', crop_direction_y='center'):
        features = []

        EI_CLASSIFIER_INPUT_WIDTH = self.dim[0]
        EI_CLASSIFIER_INPUT_HEIGHT = self.dim[1]

        in_frame_cols = img.shape[1]
        in_frame_rows = img.shape[0]

        factor_w = EI_CLASSIFIER_INPUT_WIDTH / in_frame_cols
        factor_h = EI_CLASSIFIER_INPUT_HEIGHT / in_frame_rows

        largest_factor = factor_w if factor_w > factor_h else factor_h

        resize_size_w = int(math.ceil(largest_factor * in_frame_cols))
        resize_size_h = int(math.ceil(largest_factor * in_frame_rows))
        resize_size = (resize_size_w, resize_size_h)

        resized = cv2.resize(img, resize_size, interpolation = cv2.INTER_AREA)

        if (crop_direction_x == 'center'):
            crop_x = int((resize_size_w - resize_size_h) / 2) if resize_size_w > resize_size_h else 0
        elif (crop_direction_x == 'left'):
            crop_x = 0
        elif (crop_direction_x == 'right'):
            crop_x = resize_size_w - EI_CLASSIFIER_INPUT_WIDTH
        else:
            raise Exception('Invalid value for crop_direction_x, should be center, left or right')

        if (crop_direction_y == 'center'):
            crop_y = int((resize_size_h - resize_size_w) / 2) if resize_size_h > resize_size_w else 0
        elif (crop_direction_y == 'top'):
            crop_y = 0
        elif (crop_direction_y == 'bottom'):
            crop_y = resize_size_h - EI_CLASSIFIER_INPUT_HEIGHT
        else:
            raise Exception('Invalid value for crop_direction_y, should be center, top or bottom')

        crop_region = (crop_x, crop_y, EI_CLASSIFIER_INPUT_WIDTH, EI_CLASSIFIER_INPUT_HEIGHT)

        cropped = resized[crop_region[1]:crop_region[1]+crop_region[3], crop_region[0]:crop_region[0]+crop_region[2]]

        if self.isGrayscale:
            cropped = cv2.cvtColor(cropped, cv2.COLOR_BGR2GRAY)
            pixels = np.array(cropped).flatten().tolist()

            for p in pixels:
                features.append((p << 16) + (p << 8) + p)
        else:
            pixels = np.array(cropped).flatten().tolist()

            for ix in range(0, len(pixels), 3):
                b = pixels[ix + 0]
                g = pixels[ix + 1]
                r = pixels[ix + 2]
                features.append((r << 16) + (g << 8) + b)

        return features, cropped
